#include "RaveBase/Converters/interface/CmsToRaveObjects.h"
#include "RaveBase/RaveInterface/rave/Vector6D.h"
#include "RaveBase/RaveInterface/rave/Covariance6D.h"
#include "RaveBase/RaveInterface/rave/PerigeeState5D.h"
#include "RaveBase/RaveInterface/rave/PerigeeCovariance5D.h"
#include "TrackingTools/TrajectoryParametrization/interface/CartesianTrajectoryError.h"
#include "FWCore/MessageLogger/interface/MessageLogger.h"
#include "RaveBase/Converters/interface/RaveStreamers.h"

#include "boost/cast.hpp"

#include <cmath>
#include <limits>
#include <ostream>

using namespace std;
namespace
{
rave::Track getNr ( const vector < rave::Track > & trks, int id )
{
  for ( vector< rave::Track >::const_iterator i = trks.begin();
        i != trks.end() ; ++i )
  {
    if ( i->id() == id ) return *i;
  }
  cout << "[CmsToRaveObjects] no track nr " << id << "?!?" << endl;
  exit ( 0 );
}

typedef numeric_limits<float> num;
}

rave::Track CmsToRaveObjects::convert ( const reco::TransientTrack & t, int id ) const
{
  /*
  rave::Vector6D st=convert ( t.impactPointState().globalParameters() );
  rave::Covariance6D err = convert ( t.impactPointState().cartesianParameters() );
  return rave::Track();
  */
  ostringstream ostr;
  ostr << "id" << id;
  return convert ( t.impactPointState(), t.chi2(), t.ndof(), 0, ostr.str() );
}

rave::Track CmsToRaveObjects::convert ( const TrajectoryStateOnSurface & s, float chi2,
                                        float ndf, void * p, const string & tag ) const
{
  rave::Vector6D st = convert ( s.globalParameters() );
  rave::Covariance6D err = convert ( s.cartesianError() );
  return rave::Track ( st, err, s.charge(), chi2, ndf, p, tag );
}

rave::Track CmsToRaveObjects::convert ( const FreeTrajectoryState & s ) const
{
  rave::Vector6D st = convert ( s.parameters() );
  rave::Covariance6D err = convert ( s.cartesianError() );
  return rave::Track ( st, err, s.charge(), num::quiet_NaN() /*chi2*/,
                       num::quiet_NaN () /*ndof*/, 0, "" );
}

rave::Vector6D CmsToRaveObjects::convert ( const GlobalTrajectoryParameters & p ) const
{
  return rave::Vector6D ( p.position().x(), p.position().y(), p.position().z(),
                          p.momentum().x(), p.momentum().y(), p.momentum().z() );

}

rave::Covariance3D CmsToRaveObjects::convert ( const GlobalError & e ) const
{
  return rave::Covariance3D ( e.cxx(), e.cyx(), e.czx(),
                              e.cyy(), e.czy(), e.czz() );
}

rave::Covariance6D CmsToRaveObjects::convert ( const CartesianTrajectoryError & e ) const
{
  return rave::Covariance6D ( e.matrix_old() [0][0], e.matrix_old() [0][1], e.matrix_old() [0][2],
                              e.matrix_old() [1][1], e.matrix_old() [1][2], e.matrix_old() [2][2],
                              e.matrix_old() [0][3], e.matrix_old() [0][4], e.matrix_old() [0][5],
                              e.matrix_old() [1][3], e.matrix_old() [1][4], e.matrix_old() [1][5],
                              e.matrix_old() [2][3], e.matrix_old() [2][4], e.matrix_old() [2][5],
                              e.matrix_old() [3][3], e.matrix_old() [3][4], e.matrix_old() [3][5],
                              e.matrix_old() [4][4], e.matrix_old() [4][5], e.matrix_old() [5][5] );
  /*
  return rave::Covariance6D ( e.matrix()(0,0), e.matrix()(0,1), e.matrix()(0,2),
                              e.matrix()(1,1), e.matrix()(1,2), e.matrix()(2,2),
                              e.matrix()(0,3), e.matrix()(0,4), e.matrix()(0,5),
                              e.matrix()(1,3), e.matrix()(1,4), e.matrix()(1,5),
                              e.matrix()(2,3), e.matrix()(2,4), e.matrix()(2,5),
                              e.matrix()(3,3), e.matrix()(3,4), e.matrix()(3,5),
                              e.matrix()(4,4), e.matrix()(4,5), e.matrix()(5,5) );
                              */
}

rave::Vertex CmsToRaveObjects::convert ( const TransientVertex & tv,
    const vector < rave::Track > & trks ) const
{
  vector < pair < float, rave::Track > > tracks;

  map<reco::TransientTrack, float > wm;

  if ( tv.hasTrackWeight() )
  {
    wm = tv.weightMap();
  }
  else
  {
    vector < reco::TransientTrack > trks = tv.originalTracks();
    for ( vector< reco::TransientTrack >::const_iterator i = trks.begin();
          i != trks.end() ; ++i )
    {
      wm[*i] = 1.0;
    }
  }

  for ( map<reco::TransientTrack, float >::iterator iter = wm.begin();
        iter != wm.end();
        iter++ )
  {
    // cout << "[CmsToRaveObjects] track: #" << iter->first.id() << endl;
    tracks.push_back ( pair< float, rave::Track > ( iter->second, getNr ( trks, iter->first.id() ) ) );
  }

  float ndf = tv.degreesOfFreedom();
  float chi2 = tv.totalChiSquared();
  /*
  cout << "[CmsToRaveObjects] chi2=" << chi2 << "-"
       << static_cast<float>( tv.totalChiSquared() ) << endl;
       */

  if ( ! ( tv.hasRefittedTracks() ) )
  {
    return rave::Vertex ( rave::Point3D ( tv.position().x(),
                                          tv.position().y(),
                                          tv.position().z() ),
                          CmsToRaveObjects().convert ( tv.positionError() ),
                          tracks, ndf, chi2 );
  }

  vector < pair < float, rave::Track > > ftracks;
  vector < reco::TransientTrack > rfs = tv.refittedTracks();
  vector < reco::TransientTrack > otrks = tv.originalTracks();

  if ( rfs.size() > otrks.size() )
  {
    edm::LogError ( "CmsToRaveObjects" ) << "There are more refitted tracks than original tracks! This thing will crash!";
  }

  vector < reco::TransientTrack >::const_iterator oit = otrks.begin();

  for ( vector< reco::TransientTrack >::const_iterator i = rfs.begin(); i != rfs.end() ; ++i )
  {
    /*
    cout << "[CmsToRaveObjects] refitted track: p=" << i->impactPointState().globalMomentum() << endl;
    cout << "[CmsToRaveObjects] original track: p=" << oit->impactPointState().globalMomentum() << endl;
    */
    // cout << "[CmsToRaveObjects] wm=" << wm[*i] << endl;
    ftracks.push_back ( pair< float, rave::Track > ( wm[*oit], convert ( *i ) ) );
    oit++;
  }

  return rave::Vertex ( rave::Point3D ( tv.position().x(),
                                        tv.position().y(),
                                        tv.position().z() ),
                        CmsToRaveObjects().convert ( tv.positionError() ),
                        tracks, ftracks, ndf, chi2 );
  /*
                 static_cast<float>( tv.degreesOfFreedom() ),
                 static_cast<float>( tv.totalChiSquared() ) ); */
}

rave::Vector3D CmsToRaveObjects::convert ( const GlobalPoint & p ) const
{
  return rave::Vector3D ( boost::numeric_cast<float> ( p.x() ),
                          boost::numeric_cast<float> ( p.y() ),
                          boost::numeric_cast<float> ( p.z() ) );
}

rave::Vector3D CmsToRaveObjects::convert ( const GlobalVector & v ) const
{
  return rave::Vector3D ( boost::numeric_cast<float> ( v.x() ),
                          boost::numeric_cast<float> ( v.y() ),
                          boost::numeric_cast<float> ( v.z() ) );
}

rave::PerigeeState5D CmsToRaveObjects::convert(const PerigeeTrajectoryParameters & ptp) const
{
  return rave::PerigeeState5D(ptp.transverseCurvature(), // omega
                              1./tan(ptp.theta()),          // tanLambda
                              ptp.phi(),                 // phi
                              ptp.transverseImpactParameter(), // d0
                              ptp.longitudinalImpactParameter()); // z0
}

rave::PerigeeCovariance5D CmsToRaveObjects::convert(
    const PerigeeTrajectoryError & pte, const PerigeeTrajectoryParameters & ptp) const
{
  float cotTheta = 1./tan(ptp.theta());
  AlgebraicMatrix55 jacobian;
  jacobian(0,0) = 1.;
  jacobian(1,1) = - 1. -  cotTheta*cotTheta;
  jacobian(2,2) = 1.;
  jacobian(3,3) = 1.;
  jacobian(4,4) = 1.;

  AlgebraicSymMatrix55 raveError = ROOT::Math::Similarity(jacobian, pte.covarianceMatrix());

  return rave::PerigeeCovariance5D(
      raveError(0,0), raveError(0,1), raveError(0,2), raveError(0,3), raveError(0,4),
      raveError(1,1), raveError(1,2), raveError(1,3), raveError(1,4), raveError(2,2),
      raveError(2,3), raveError(2,4), raveError(3,3), raveError(3,4), raveError(4,4) );
}

#ifdef WITH_KINEMATICS

rave::Vector7D CmsToRaveObjects::convert ( const ::KinematicParameters & p ) const
{
  return rave::Vector7D ( CmsToRaveObjects().convert ( p.position() ),
                          CmsToRaveObjects().convert ( p.momentum() ),
                          p.vector() ( 6 ) );
}

rave::Covariance7D CmsToRaveObjects::convert ( const ::KinematicParametersError & e ) const
{
  return rave::Covariance7D (
           e.matrix() (0,0), e.matrix() (0,1), e.matrix() (0,2),
           e.matrix() (1,1), e.matrix() (1,2), e.matrix() (2,2),
           e.matrix() (0,3), e.matrix() (0,4), e.matrix() (0,5),
           e.matrix() (1,3), e.matrix() (1,4), e.matrix() (1,5),
           e.matrix() (2,3), e.matrix() (2,4), e.matrix() (2,5),
           e.matrix() (3,3), e.matrix() (3,4), e.matrix() (3,5),
           e.matrix() (4,4), e.matrix() (4,5), e.matrix() (5,5),
           e.matrix() (0,6), e.matrix() (1,6), e.matrix() (2,6),
           e.matrix() (3,6), e.matrix() (4,6), e.matrix() (5,6),
           e.matrix() (6,6) );
           /* e.matrix() [0][0], e.matrix() [0][1], e.matrix() [0][2],
           e.matrix() [1][1], e.matrix() [1][2], e.matrix() [2][2],
           e.matrix() [0][3], e.matrix() [0][4], e.matrix() [0][5],
           e.matrix() [1][3], e.matrix() [1][4], e.matrix() [1][5],
           e.matrix() [2][3], e.matrix() [2][4], e.matrix() [2][5],
           e.matrix() [3][3], e.matrix() [3][4], e.matrix() [3][5],
           e.matrix() [4][4], e.matrix() [4][5], e.matrix() [5][5],
           e.matrix() [0][6], e.matrix() [1][6], e.matrix() [2][6],
           e.matrix() [3][6], e.matrix() [4][6], e.matrix() [5][6],
           e.matrix() [6][6] ); */
}

#endif
